//-----------------------------------------------------------------------------
// MK3_2560_FAT16
// Version 2.1
// 02.11.2011 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board
// mit ATMega-2560 @ 16MHz
// SD-Karte Ein-/Ausgabe (FAT16-Fileformat)
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_FAT16.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : Port-B (Bit0)
// Benutzt : MK3_2560_SPI.h = SPI = Port-B (Bit1 bis Bit3)
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Copyright Hinweis von UB : 
// ich habe den Code fuer das FAT16-Dateisystem von dieser Quelle :
// Adr : www.Mikro-control.de
// File : MikroFat16
// Ver : 1.29
// Autor : Stephan Busker
//-----------------------------------------------------------------------------

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// + Copyright (c) 2008 Stephan Busker
// + Nur fr den privaten Gebrauch
// + FOR NON COMMERCIAL USE ONLY
// + www.Mikro-control.de
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// + Es gilt fr das gesamte Projekt (Hardware, Software, Binrfiles, Sourcecode und Dokumentation), 
// + dass eine Nutzung (auch auszugsweise) nur fr den privaten (nicht-kommerziellen) Gebrauch zulssig ist. 
// + Sollten direkte oder indirekte kommerzielle Absichten verfolgt werden, ist mit uns (stephan.busker@mikro-control.de) Kontakt 
// + bzgl. der Nutzungsbedingungen aufzunehmen. 
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// + Keine Gewhr auf Fehlerfreiheit, Vollstndigkeit oder Funktion
// + Benutzung auf eigene Gefahr
// + Wir bernehmen keinerlei Haftung fr direkte oder indirekte Personen- oder Sachschden
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// + Redistributions of source code (with or without modifications) must retain the above copyright notice, 
// + this list of conditions and the following disclaimer.
// +   * Neither the name of the copyright holders nor the names of contributors may be used to endorse or promote products derived
// +     from this software without specific prior written permission.
// +   * The use of this project (hardware, software, binary files, sources and documentation) is only permitted 
// +     for non-commercial use (directly or indirectly)
// +     Commercial use is only permitted with our written permission
//
// +  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// +  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// +  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// +  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// +  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// +  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// +  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// +  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// +  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// +  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// +  POSSIBILITY OF SUCH DAMAGE. 
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++





//----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include "ub_mk3_2560/MK3_2560_SPI.h" //  bindet die SPI-Funktionen ein



//-----------------------------------------------------------------------------
// Hier koennen Funktionen die nicht benutzt werden
// abgeschaltet werden...einfach den DEFINE von "1" auf "0" aendern
//-----------------------------------------------------------------------------

#define USE_MK3_F_OPENCLOSE 1   // fuer F21+F22
#define USE_MK3_F_READ 1        // fuer F23+F29+F52
#define USE_MK3_F_WRITE 1       // fuer F24+F28
#define USE_MK3_F_SEARCH 1      // fuer F30
#define USE_MK3_F_DIR 1         // fuer F45+F46
#define USE_MK3_F_FIND 1        // fuer F48+F49
#define USE_MK3_F_DELETE 1      // fuer F50+F51



//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "MK3_INIT_FAT16" starten
// 02. Funktion "MK3_FAT16_CHKDSK" starten
// 03. Wenn Rueckgabe von 02.=TRUE weitermachen, sonst -> keine SD-Karte gefunden


//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
// MK3_INIT_FAT16()     // init Funktion der FAT
// MK3_FAT16_CHKDSK()   // Check ob eine SD-Karte vorhanden ist
// F21_fopen_()         // oeffnet ein File (zum lesen oder schreiben)
// F22_fclose_()        // schliesse ein File
// F23_fread_()         // String lesen aus einem File
// F24_fwrite_()        // String schreiben in ein File
// F28_fputs_()         // Zeichen schreiben in ein File
// F29_fgets_()         // Zeichen lesen aus einem File
// F30_fexist_()        // test ob ein File existiert
// F45_mkdir_()         // erstellt ein Directory
// F46_chdir_()         // wechselt in ein Directory
// F48_findfirst_()     // findet das erste Objekt
// F49_findnext_()      // findet das naechste Objekt
// F50_fdelete_()       // loeschet ein File
// F51_rmdir_()         // loeschet ein Directory
// F52_feof_()          // test ob File-Ende erreicht
//-----------------------------------------------------------------------------



//----------------------------------------------------------------------------
// Typen-Deklaration----------------------------------------------------------
//----------------------------------------------------------------------------
typedef unsigned char  u8;
typedef   char  s8;
typedef unsigned short u16;
typedef   signed short s16;
typedef unsigned long  u32;
typedef   signed long  s32;
typedef const char cs8;



//----------------------------------------------------------------------------
// ENUM-Definition------------------------------------------------------------
//----------------------------------------------------------------------------
typedef enum
{
  SD_SUCCESS = 0,
  SD_ERROR_NOCARD,
  SD_ERROR_RESET,
  SD_ERROR_INITIALIZE,
  SD_ERROR_BAD_RESPONSE,
  SD_ERROR_BAD_VOLTAGE_RANGE,
  SD_ERROR_NO_SDCARD,
  SD_ERROR_TIMEOUT,
  SD_ERROR_CRC_DATA,
  SD_ERROR_WRITE_DATA,
  SD_ERROR_READ_DATA,
  SD_ERROR_SET_BLOCKLEN,
  SD_ERROR_UNKNOWN
} SD_Result_t;


typedef enum
{
	VER_UNKNOWN,
 	VER_1X,
	VER_20
} SDVersion_t;

typedef enum
{
	O_SEEK_SET,
	O_SEEK_CUR,
	O_SEEK_END
} Origin_t;

//----------------------------------------------------------------------------
// Struktur-Definition--------------------------------------------------------
//----------------------------------------------------------------------------
typedef struct afile
{
	u32 start_cluster;				
	u32 cluster_pointer;			
	u8 sector_index;				
	u16 byte_index;					
	u8 mode;						
	u32 filesize;					
	u32 fileposition;				
	u32 sector_in_buffer;			
	u32 directory_root_sector;		
	u32 directory_sector;			
	u8 directory_index;			
	u8 attribute;					
	u8  buffer[512];				
	u8 state;						
} File;

typedef struct find
{
	s8 name[12];					
	s8 searchstring[13];
	u32 cluster_pointer;			
	u16 startcluster;				
	u32 filesize;					
	u32 directory_sector;			
	u8 directory_index;			
	u8 attribute;					
} Find;

struct DirEntry
{
	u8   name[8];					
	u8   extension[3];				
	u8   attribute;					
	u8   reserved[10];				
	u16  time;						
	u16  date;						
	u16  startcluster;				
	u32  size;						
//} __attribute__((packed));
};

struct FatEntry
{
	u16  next_cluster;				
//}  __attribute__((packed)); 
};

struct PartitionEntry
{
	u8	PartitionState;
	u8	BeginningHead;
	u16	BeginningCylinder;
	u8	Type;
	u8	EndHead;
	u16	EndCylinder;
	u32	NoSectorsBeforePartition;
	u32	NoSectorsPartition	;
//} __attribute__((packed)); 
};

struct VBR_Entry
{
	u8  dummy[11];						
	u16 bps;
	u8  SectorsPerCluster;
	u16 ReservedSectors;
	u8  NoFATCopies;
	u16 MaxRootEntries;
	u16 dummy2;
	u8  dummy3;
	u16 SectorsPerFAT;
//} __attribute__((packed)); 
};

struct MBR_Entry
{
	u8  ExecutableCode[446];			
	struct 	PartitionEntry	PartitionEntry1;	
	struct 	PartitionEntry	PartitionEntry2;	
	struct 	PartitionEntry	PartitionEntry3;	
	struct 	PartitionEntry	PartitionEntry4;	
	u16		ExecutableMarker;	
//} __attribute__((packed)); 
};

typedef struct
{
  u8 Valid;
  SDVersion_t Version;  // HW-Version 
  u32 Capacity;		// Memory capacity  in bytes
  u8 CID[16];		// CID register
  u8 CSD[16];		// CSD register
//} __attribute__((packed)) SDCardInfo_t;
}SDCardInfo_t;

volatile SDCardInfo_t SDCardInfo;

struct DirEntry 	*DirectoryEntry;	
struct FatEntry 	*Fat;			

//----------------------------------------------------------------------------
// Prototypen-----------------------------------------------------------------
// von EXTERNEN-Funktionen----------------------------------------------------
//----------------------------------------------------------------------------
extern SD_Result_t	F10_SDC_Init(void);
extern SD_Result_t 	F11_SDC_PutSector (u32, const u8 *);
extern SD_Result_t 	F12_SDC_GetSector (u32 ,u8 *);

extern void   F13_SSC_Init(void);
extern void   F15_SSC_ClearRxFifo(void);
extern u8     F16_SSC_GetChar (void);
extern void   F17_SSC_PutChar (u8);
extern void   F18_SSC_Disable(void);
extern void   F19_SSC_Enable(void);

void 		MK3_INIT_FAT16(void);
extern bool	MK3_FAT16_CHKDSK(void);		
extern File *	F21_fopen_(cs8 *fname,s8 mode);
extern void 	F22_fclose_(File *file);
extern u32 	F23_fread_(void *buffer, u32 size, u32 count, File *file); 
extern u32 	F24_fwrite_(void *buffer, u32 size, u32 count, File *file);
extern s16  	F25_fseek_(File *file, s32 offset, Origin_t origin);
extern s16 	F26_fgetchar_(File *file);
extern u8	F27_fputchar_(File *file,s8 c);
extern u8	F28_fputs_(File *file,cs8 *string);
extern s8 *  	F29_fgets_(s8 *s, s16 count, File *file);
extern u8 	F30_fexist_(cs8 *fname);
extern u16   	F31_GetNextCluster(File *file);
extern u16 	F32_FindNextFreeCluster(File *file);
extern u8 	F33_AppendCluster(File *file);
extern void     F34_DeleteClusterChain(u16 startcluster);
extern u16 	F35_GetFatClusterOffset(File *file);
extern u16 	F36_GetFatSectorIndex(File *file);
extern u8	F37_CreateFileInDirectory(cs8 *fname, File *file);
extern u8	F38_SeekFileInDirectory(cs8 *fname, File *file);
extern void 	F39_SeperateFileName(cs8 *fname, s8 *name);
extern File * 	F40_ReserveFilePointer(void);
extern void 	F41_FreeFilePointer(File *file);
extern void     F42_DeleteDirectoryEntry(Find *item);
extern u8 	F43_CreateSubDirectory_(cs8 *fname);
extern u16 	F44_SeekSubDirectory(cs8 *fname);
extern u8	F45_mkdir_(cs8 *fname);
extern u8	F46_chdir_(cs8 *fname);
extern u8	F47_FindItem(Find *item);
extern u8	F48_findfirst_(cs8 *fname, Find *item, u8 attribute);
extern u8	F49_findnext_(Find *item);
extern u8       F50_fdelete_(cs8 *fname);
extern u8       F51_rmdir_(cs8 *dname);
extern u16	F52_feof_(File *);

//----------------------------------------------------------------------------
// Prototypen-----------------------------------------------------------------
// von INTERNEN-Funktionen----------------------------------------------------
//----------------------------------------------------------------------------

u16 I01_CRC16(const u8 * pBuffer, u32 len);
u8 I02_CRC7(u8 * cmd, u32 len);
u8 I03_SDC_WaitForBusy(void);
u8 I04_SDC_SendCMDR1(u8 CmdNo, u32 arg);
u8 I05_SDC_SendACMDR1(u8 CmdNo, u32 arg);
SD_Result_t I06_SDC_GetData(u8 CmdNo, u32 addr, u8 *Buffer, u32 len);
void I07_SDC_PrintCID(u8 * pCID);
SD_Result_t I08_SDC_GetCID(u8 * pCID);
SD_Result_t I09_SDC_GetCSD(u8 * pCSD);
s16 I10_fflush_(File *file);

//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------

#define MK3_DDR_SD DDRB		// DDR-Register der SD-Karte
#define MK3_PORT_SD PORTB	// Port der SD-Karte

#define MK3_SD_SlaveSel 0x01    // SlaveSelect von SD

//----------------------------------------------------------------------------

#define		__MAX_FILES_USED	1	

#define		_UNUSED		1		
#define		_USED		2
#define		_ARCHIVE	2
#define		_READ_ONLY	4
#define		_SYSTEM		8
#define		_DIRECTORY	16
#define		_FILE		32


#define	_EMPTY						0x00   
#define	_FAT12						0x01 
#define	_FAT16_ST_32_MB					0x04 
#define	_EXTENDED					0x05 
#define	_FAT16_LT_32_MB					0x06 
#define	_HPFS						0x07 
#define	_FAT32 						0x0B 
#define	_FAT32_BIOS_Extension 			        0x0C 
#define	_FAT16_32_MB_BIOS_Extension		        0x0E 
#define	_EXTENDED_BIOS_Extension		        0x0F 
#define	_EISA						0x12 
#define	_DYNAMIC					0x42 
#define	_Linux_Swap					0x82 
#define	_Linux_Native 					0x83 
#define	_Linux_LVM 					0x8E 
#define	_FreeBSD 					0xA5 
#define	_OpenBSD 					0xA6 
#define	_NetBSD 					0xA9 

#define		_MBR_SECTOR	0		

#define SDC_DEBUG
#define CARD_DETECT (!(PINB & 0x08))

#define CMD_GO_IDLE_STATE		0x00	
#define CMD_SEND_OP_COND		0x01 	
#define CMD_SEND_IF_COND		0x08	
#define CMD_SEND_CSD			0x09	
#define CMD_SEND_CID			0x0A 	
#define CMD_SEND_STATUS			0x0D	
#define CMD_SET_BLOCKLEN		0x10	
#define CMD_READ_SINGLE_BLOCK 	        0x11	
#define CMD_WRITE_SINGLE_BLOCK	        0x18	
#define CMD_APP_CMD			0x37	
#define CMD_READ_OCR 			0x3A 	
#define CMD_CRC_ON_OFF 			0x3B 	
#define ACMD_SEND_OP_COND		0x29	

#define R1_NO_ERR	 		0x00
#define R1_IDLE_STATE 			0x01
#define R1_ERASE_RESET 			0x02
#define R1_ILLEGAL_CMD 			0x04
#define R1_COM_CRC_ERR	 		0x08
#define R1_ERASE_SEQUENCE_ERR	        0x10
#define R1_ADDRESS_ERR 			0x20
#define R1_PARAMETER_ERR 		0x40
#define R1_BAD_RESPONSE 		0x80

#define R2_NO_ERR	 		0x00
#define R2_CARD_LOCKED 			0x01
#define R2_ERASE_WRITE_PROT_ERR	        0x02
#define R2_UNKOWN_ERR			0x04
#define R2_CARD_CTRL_ERR 		0x08
#define R2_CARD_ECC_ERR		        0x10
#define R2_WRITE_PROT_ERR		0x20
#define R2_ERASE_PARAM_ERR 		0x40
#define R2_OUT_OF_RANGE_ERR		0x80

#define DATA_START_TOKEN 		0xFE
#define DATA_RESPONSE_MASK 		0x1F 
#define DATA_RESPONSE_OK 		0x05
#define DATA_RESPONSE_CRC_ERR	        0x0B
#define DATA_RESPONSE_WRITE_ERR         0x1D	



//----------------------------------------------------------------------------
// Variabeln------------------------------------------------------------------
//----------------------------------------------------------------------------
extern u8		SectorsPerCluster =0 ; // AVR-Studio Compiler-Warning ist normal
extern u32		CWD=0; // AVR-Studio Compiler-Warning ist normal
char text[50];

u8		FatCopies 			= 0;	
u16		PossibleRootEntries 	        = 0;			
u16		SectorsPerFat 			= 0;		
u32		ReservedSectors 		= 0;	
u32		FirstPartitionSector 	        = 0;		
u32		FileAllocationTable 	        = 0;		
u32		RootDirectory 			= 0;	
u32		FirstDataCluster 		= 0;


File FilePointer[__MAX_FILES_USED];		



//----------------------------------------------------------------------------
// Konstante------------------------------------------------------------------
//----------------------------------------------------------------------------

static const u16 crc16tab[256]= 
{
	0x0000,0x1021,0x2042,0x3063,0x4084,0x50a5,0x60c6,0x70e7,
	0x8108,0x9129,0xa14a,0xb16b,0xc18c,0xd1ad,0xe1ce,0xf1ef,
	0x1231,0x0210,0x3273,0x2252,0x52b5,0x4294,0x72f7,0x62d6,
	0x9339,0x8318,0xb37b,0xa35a,0xd3bd,0xc39c,0xf3ff,0xe3de,
	0x2462,0x3443,0x0420,0x1401,0x64e6,0x74c7,0x44a4,0x5485,
	0xa56a,0xb54b,0x8528,0x9509,0xe5ee,0xf5cf,0xc5ac,0xd58d,
	0x3653,0x2672,0x1611,0x0630,0x76d7,0x66f6,0x5695,0x46b4,
	0xb75b,0xa77a,0x9719,0x8738,0xf7df,0xe7fe,0xd79d,0xc7bc,
	0x48c4,0x58e5,0x6886,0x78a7,0x0840,0x1861,0x2802,0x3823,
	0xc9cc,0xd9ed,0xe98e,0xf9af,0x8948,0x9969,0xa90a,0xb92b,
	0x5af5,0x4ad4,0x7ab7,0x6a96,0x1a71,0x0a50,0x3a33,0x2a12,
	0xdbfd,0xcbdc,0xfbbf,0xeb9e,0x9b79,0x8b58,0xbb3b,0xab1a,
	0x6ca6,0x7c87,0x4ce4,0x5cc5,0x2c22,0x3c03,0x0c60,0x1c41,
	0xedae,0xfd8f,0xcdec,0xddcd,0xad2a,0xbd0b,0x8d68,0x9d49,
	0x7e97,0x6eb6,0x5ed5,0x4ef4,0x3e13,0x2e32,0x1e51,0x0e70,
	0xff9f,0xefbe,0xdfdd,0xcffc,0xbf1b,0xaf3a,0x9f59,0x8f78,
	0x9188,0x81a9,0xb1ca,0xa1eb,0xd10c,0xc12d,0xf14e,0xe16f,
	0x1080,0x00a1,0x30c2,0x20e3,0x5004,0x4025,0x7046,0x6067,
	0x83b9,0x9398,0xa3fb,0xb3da,0xc33d,0xd31c,0xe37f,0xf35e,
	0x02b1,0x1290,0x22f3,0x32d2,0x4235,0x5214,0x6277,0x7256,
	0xb5ea,0xa5cb,0x95a8,0x8589,0xf56e,0xe54f,0xd52c,0xc50d,
	0x34e2,0x24c3,0x14a0,0x0481,0x7466,0x6447,0x5424,0x4405,
	0xa7db,0xb7fa,0x8799,0x97b8,0xe75f,0xf77e,0xc71d,0xd73c,
	0x26d3,0x36f2,0x0691,0x16b0,0x6657,0x7676,0x4615,0x5634,
	0xd94c,0xc96d,0xf90e,0xe92f,0x99c8,0x89e9,0xb98a,0xa9ab,
	0x5844,0x4865,0x7806,0x6827,0x18c0,0x08e1,0x3882,0x28a3,
	0xcb7d,0xdb5c,0xeb3f,0xfb1e,0x8bf9,0x9bd8,0xabbb,0xbb9a,
	0x4a75,0x5a54,0x6a37,0x7a16,0x0af1,0x1ad0,0x2ab3,0x3a92,
	0xfd2e,0xed0f,0xdd6c,0xcd4d,0xbdaa,0xad8b,0x9de8,0x8dc9,
	0x7c26,0x6c07,0x5c64,0x4c45,0x3ca2,0x2c83,0x1ce0,0x0cc1,
	0xef1f,0xff3e,0xcf5d,0xdf7c,0xaf9b,0xbfba,0x8fd9,0x9ff8,
	0x6e17,0x7e36,0x4e55,0x5e74,0x2e93,0x3eb2,0x0ed1,0x1ef0
};
  

//----------------------------------------------------------------------------
// Funktionen-----------------------------------------------------------------
//----------------------------------------------------------------------------

u16 I01_CRC16(const u8 *pBuffer, u32 len)
{
	register u32 counter;
	register u16 crc = 0;
	for( counter = 0; counter < len; counter++)
		crc = (crc<<8) ^ crc16tab[((crc>>8) ^ *pBuffer++)&0x00FF];
	return crc;
}

u8 I02_CRC7(u8 * cmd, u32 len)
{
	u8 i, a;
	u8 crc, Data;

	crc = 0; // init CRC buffer
	for (a = 0; a < len ;a++) // for every byte in the msg
	{
		Data = cmd[a];
		for (i=0;i<8;i++) // for every bit in the byte
		{
			crc <<= 1; // shift crc
			if ((Data & 0x80)^(crc & 0x80))	crc ^=0x09;	
			Data <<= 1;	// shift data  for next bit
		}
	}
	crc = (crc<<1)|1; // set terminating bit to 1
	return(crc);
}

u8 I03_SDC_WaitForBusy(void)
{
	u8 rsp = 0;
	u16 timeout=0;

	F15_SSC_ClearRxFifo();
	F19_SSC_Enable();				
	do	
	{
		rsp = F16_SSC_GetChar();
		if(timeout++>500) break;
	}while(rsp != 0xFF);		
	return(rsp);
}

u8 I04_SDC_SendCMDR1(u8 CmdNo, u32 arg)
{
	u8 r1;
	u16 Timeout = 0;
	u16 a;
	u8 cmd[6];

	F15_SSC_ClearRxFifo();	// clear the rx fifo
	F19_SSC_Enable();		// enable chipselect.
	I03_SDC_WaitForBusy();
	F15_SSC_ClearRxFifo();	// clear the rx fifo
	F16_SSC_GetChar();      // dummy to sync

	/* Send cmd10 (SEND_CID) */;
	cmd[0] = 0x40|CmdNo; // set command index
	cmd[1] = (arg & 0xFF000000)>>24;
	cmd[2] = (arg & 0x00FF0000)>>16;
	cmd[3] = (arg & 0x0000FF00)>>8;
   	cmd[4] = (arg & 0x000000FF);
	cmd[5] = I02_CRC7(cmd, 5); // update checksum 
	for (a = 0;a < 6; a++) 
	{	
		F17_SSC_PutChar(cmd[a]);
	}
	F15_SSC_ClearRxFifo();	
	do					
	{
		r1 = F16_SSC_GetChar();       // get byte from sd-card
		if (Timeout++ > 500) return(r1);
	}while(r1 == 0xFF); // wait for the response byte from sd-card.
	return(r1);
}

u8 I05_SDC_SendACMDR1(u8 CmdNo, u32 arg)
{
  	u8 r1 = 0xFF;
	r1 = I04_SDC_SendCMDR1(CMD_APP_CMD, 0UL);
	if(r1 & R1_BAD_RESPONSE) return(r1);
	r1 = I04_SDC_SendCMDR1(CmdNo, arg);
	return(r1);
}

SD_Result_t I06_SDC_GetData(u8 CmdNo, u32 addr, u8 *Buffer, u32 len)
{
	u8 rsp; 	
	u16 a, Crc16;
	SD_Result_t result = SD_ERROR_UNKNOWN;

	// send the command	
	rsp = I04_SDC_SendCMDR1(CmdNo, addr);
	if (rsp != R1_NO_ERR) 
	{
		result = SD_ERROR_BAD_RESPONSE;
		goto end;
	}
	F15_SSC_ClearRxFifo();
	do
	{
		rsp = F16_SSC_GetChar();
		if((rsp & 0xF0) == 0x00) // data error token 
		{
			result = SD_ERROR_READ_DATA;
			goto end;
		}
//		if(timeout++>500) break;
	}while(rsp != DATA_START_TOKEN);
	// data start token received
	for (a = 0; a < len; a++)	// read the block from the SSC
	{
		Buffer[a] = F16_SSC_GetChar();
	}
	// Read two bytes CRC16-Data checksum
	Crc16 = F16_SSC_GetChar(); // highbyte fisrt	
	Crc16 = (Crc16<<8)|F16_SSC_GetChar(); // lowbyte last
	result = SD_SUCCESS;
	
	end:
	return(result);
}

void I07_SDC_PrintCID(u8 * pCID)
{
/*	u8 pn[6];
	u16 temp1, temp2;

	sprintf(&text[0], "\r\nManufacturer ID: %i\r\n", pCID[0]);
	PutString(text);
	memcpy(pn, &pCID[1], 2);
	pn[2] = '\0'; // terminate string
	sprintf(&text[0], "Application ID: %s\r\n",pn);
	PutString(text);
	memcpy(pn, &pCID[3], 5);
	pn[5] = '\0'; // terminate string
	sprintf(&text[0], "Product Name: %s\r\n",pn);
	PutString(text);
	sprintf(&text[0], "Product Rev.: %i.%i\r\n",pCID[8]>>4, pCID[8]&0xF);
	PutString(text);
	PutString("Serial No.: ");
	for(temp1 = 0; temp1<4; temp1++)
	{
		sprintf(&text[0],"%02X", pCID[9+temp1]);
		PutString(text);
	}
	PutString("\r\n");
	temp1 = pCID[14] & 0x0F;    // month 
	temp2 = ((pCID[14]>>4)|(pCID[13]<<4)) + 2000; // year 
	sprintf(&text[0], "Manufac. Date: %i/%i\r\n\r\n",temp1, temp2);
	PutString(text);
*/
}

SD_Result_t I08_SDC_GetCID(u8 * pCID) 
{
	return I06_SDC_GetData(CMD_SEND_CID, 0UL, pCID, 16);
}


SD_Result_t I09_SDC_GetCSD(u8 * pCSD) 
{
	return I06_SDC_GetData(CMD_SEND_CSD, 0UL, pCSD, 16);
}

s16	I10_fflush_(File *file)
{
	u16 time=0;
	u16 date=0;

			
	if(file && ((file->mode =='a') || (file->mode =='w')))
	{
		if(file->byte_index > 0)	
		{
			F11_SDC_PutSector((u32)(file->cluster_pointer + file->sector_index),file->buffer);	
		}	
		F12_SDC_GetSector((u32)file->directory_sector,file->buffer);
		DirectoryEntry = (struct DirEntry *)file->buffer;
		DirectoryEntry[file->directory_index].size = (u32) file->filesize;
		DirectoryEntry[file->directory_index].time = (u16) time;
		DirectoryEntry[file->directory_index].date = (u16) date;
		F11_SDC_PutSector((u32)file->directory_sector,file->buffer);
	}
	return(0);
}



SD_Result_t F10_SDC_Init(void)
{
	u32 Timeout = 0;
	u8 rsp[6]; // SD-SPI response buffer
	SD_Result_t result = SD_ERROR_UNKNOWN;     

    	F13_SSC_Init();

	SDCardInfo.Valid = 0;

	F18_SSC_Disable(); // set SD_CS high
	F15_SSC_ClearRxFifo();	// clear the rx fifo


	
	for (Timeout = 0; Timeout < 15; Timeout++) 	
	{
		F17_SSC_PutChar(0xFF);
	}

	
	// switch to idle state
	while(!(I04_SDC_SendCMDR1(CMD_GO_IDLE_STATE, 0UL) & R1_IDLE_STATE))
	{
		if (Timeout++ > 20)
		{
			result = SD_ERROR_RESET;
			goto end;
		}
	}

	rsp[0] = I04_SDC_SendCMDR1(CMD_SEND_IF_COND, 0x000001AA);
	// answer to cmd58 is an R7 response (R1+ 4Byte IFCond)
	if(rsp[0] & R1_BAD_RESPONSE)
	{
		result = SD_ERROR_BAD_RESPONSE;
		goto end;
	} 
	if(rsp[0] & R1_ILLEGAL_CMD)
	{
		//Ver1.X SD Memory Card or not a SD Memory Card
		SDCardInfo.Version = VER_1X;			
	}
	else
	{
		// Ver2.00 or later SD Memory Card
		// reading the remaining bytes of the R7 response
		SDCardInfo.Version = VER_20;
		for(Timeout = 1; Timeout < 5; Timeout++)
		{
			rsp[Timeout] = F16_SSC_GetChar();
		}
		//check pattern
		if(rsp[4]!= 0xAA)
		{
			result = SD_ERROR_BAD_RESPONSE;
			goto end;
		}
		if ( (rsp[3] & 0x0F)!= 0x01 )
		{
			result = SD_ERROR_BAD_VOLTAGE_RANGE;
			goto end;		
		} 
	}
	
	rsp[0] = I04_SDC_SendCMDR1(CMD_READ_OCR, 0UL);
	// answer to cmd58 is an R3 response (R1 + 4Byte OCR)
	if(rsp[0] & R1_BAD_RESPONSE)
	{
		result = SD_ERROR_BAD_RESPONSE;
		goto end;
	}
	if(rsp[0] & R1_ILLEGAL_CMD)
	{
		result = SD_ERROR_NO_SDCARD;
		goto end;
	}
	// read 4 bytes of OCR register
	for(Timeout = 1; Timeout < 5; Timeout++)
	{
		rsp[Timeout] = F16_SSC_GetChar();
	}
	//	NavicCtrl uses 3.3 V,  therefore check for bit 20 & 21 
	if((rsp[2] & 0x30) != 0x30)
	{
		// supply voltage is not supported by sd-card
		result = SD_ERROR_BAD_VOLTAGE_RANGE;
		goto end; 
	}
		
	// Initialize the sd-card s

	do
	{
		rsp[0] = I05_SDC_SendACMDR1(ACMD_SEND_OP_COND, 0UL);
		if(rsp[0] & R1_BAD_RESPONSE)
		{
			result = SD_ERROR_BAD_RESPONSE;
			goto end;
		}
	} while(rsp[0] & R1_IDLE_STATE); // loop until idle state
		
	if(rsp[0] != R1_NO_ERR)
	{
		 result = SD_ERROR_INITIALIZE;
		goto end; 	
	}
	
    	if(I04_SDC_SendCMDR1(CMD_SET_BLOCKLEN, 512UL) != R1_NO_ERR)
    	{
		 result = SD_ERROR_SET_BLOCKLEN;
		goto end;
    	}

	// read CID register
	result = I08_SDC_GetCID((u8 *)&SDCardInfo.CID);
	if(result != SD_SUCCESS)
	{
		goto end;
	}
	
	// read CSD register
	result = I09_SDC_GetCSD((u8 *)&SDCardInfo.CSD);
	if(result != SD_SUCCESS)
	{
		goto end;
	}
	
	u8 c_size_mult, read_bl_len;
	u32 c_size;
	
	switch(SDCardInfo.CSD[0]>>6) // check CSD Version
	{
	case 0x00: // if CSD is V1.0 structure (2GB limit)
	
	
		read_bl_len = (SDCardInfo.CSD[5] & 0x0F); 		
		c_size = ((u32)(SDCardInfo.CSD[6] & 0x03))<<10;
		c_size |= ((u32)SDCardInfo.CSD[7])<<2;			
		c_size |= (u32)(SDCardInfo.CSD[8]>>6);			
		c_size_mult = (SDCardInfo.CSD[9] & 0x03)<<1;  	
		c_size_mult |=(SDCardInfo.CSD[10] & 0x80)>>7;	
		SDCardInfo.Capacity = (u32)(c_size+1)*(1L<<(c_size_mult+2))*(1L<<read_bl_len);
		break;
	
	case 0x01: // if CSD is V2.0 structure (HC SD-Card > 2GB) 
	
	
		c_size = ((u32)(SDCardInfo.CSD[7] & 0x3F))<<16;
		c_size |= ((u32)SDCardInfo.CSD[8])<<8;
		c_size |= (u32)SDCardInfo.CSD[9];
		 SDCardInfo.Capacity = (c_size + 1)* 512L * 1024L;
		break;
	
	default: //unknown CSD Version
		SDCardInfo.Capacity = 0;
		break;	
	}


	
//	u16 mb_size = (u16)(SDCardInfo.Capacity/(1024L*1024L));	
//	s8 text[40];
//	sprintf(text,"\r\nsize: %d",mb_size);
//	PutString(text);
//	F07_SDC_PrintCID((u8 *)&SDCardInfo.CID);
	SDCardInfo.Valid = 1;
			
	// jump point for error	condition before
	end:
	F18_SSC_Disable();


	return(result);
}

SD_Result_t F11_SDC_PutSector(u32 addr, const u8 *Buffer)
{
	u8 rsp; 
        char uok=0;
	
	u16 timeout = 0;
	u16 a, Crc16;
	//u32 Timeout = 0;
	SD_Result_t result = SD_ERROR_UNKNOWN;

	addr = addr << 9; // convert sectoradress to byteadress
	rsp = I04_SDC_SendCMDR1(CMD_WRITE_SINGLE_BLOCK, addr);
	if (rsp != R1_NO_ERR)
	{
		result = SD_ERROR_BAD_RESPONSE;
		goto end;
	}
	F15_SSC_ClearRxFifo();		
	for (a=0;a<20;a++)					
	{
		F16_SSC_GetChar();
	}
	Crc16 = I01_CRC16(Buffer, 512);        
	F17_SSC_PutChar(DATA_START_TOKEN);		
	
	for (a=0;a<512;a++)					
	{
		F17_SSC_PutChar(Buffer[a]);
	}
	// write two bytes of crc16 to the sdcard
	F17_SSC_PutChar((u8)(Crc16>>8)); 		
	F17_SSC_PutChar((u8)(0x00FF&Crc16)); 	
	F15_SSC_ClearRxFifo();
	do
	{
	 	rsp = F16_SSC_GetChar();
		if(timeout++>500) break;
	}while((rsp & 0x11) != 0x01 );
	// analyse data response token
	switch(rsp & DATA_RESPONSE_MASK)
	{
		case DATA_RESPONSE_OK:
			result = SD_SUCCESS;
			break;
		case DATA_RESPONSE_CRC_ERR:
			result = SD_ERROR_CRC_DATA;
			uok=1;
			break;
		case DATA_RESPONSE_WRITE_ERR:
			result = SD_ERROR_WRITE_DATA;
			uok=1;
			break;
		default:
			result = SD_ERROR_UNKNOWN;
			uok=1;
			break;

	}
        if(uok==0) {
	  // wait until the sdcard is busy.
	  rsp = I03_SDC_WaitForBusy();
	  if(rsp != 0xFF)
	  {
	 	result =  SD_ERROR_TIMEOUT;
	 	goto end;
	  }
	  // check card status
	  rsp = I04_SDC_SendCMDR1(CMD_SEND_STATUS, 0);
	  if(rsp != R1_NO_ERR)
	  {
		result =  SD_ERROR_BAD_RESPONSE;
		F16_SSC_GetChar();
	 	goto end;
	  }
	  // 2nd byte of r2 response
	  rsp = F16_SSC_GetChar();
	  if(rsp != R2_NO_ERR)
	  {
		result =  SD_ERROR_WRITE_DATA;
		F16_SSC_GetChar();
	 	goto end;
	  }
        }
        end:
	//F18_SSC_Disable();
	return(result);
} 

SD_Result_t F12_SDC_GetSector(u32 addr,u8 *Buffer)
{
	addr = addr << 9; // convert sectoradress to byteadress
	return I06_SDC_GetData(CMD_READ_SINGLE_BLOCK, addr, Buffer, 512);
}


void F13_SSC_Init(void)
{

  F18_SSC_Disable();	

}


void F15_SSC_ClearRxFifo(void)
{
	// enter your code here to clear the rx-fifo of the ssc.	
}

u8 F16_SSC_GetChar (void)
{
	u8 Byte = 0;

	Byte=MK3_SPI_SEND(0xFF);

	return Byte;

}

void F17_SSC_PutChar (u8 Byte)
{

	MK3_SPI_SEND(Byte);
}

void F18_SSC_Disable(void) 
{ 
  MK3_PORT_SD|=MK3_SD_SlaveSel;  // SlaveSelect AUS
}

void F19_SSC_Enable(void) 
{ 
  MK3_PORT_SD&=~MK3_SD_SlaveSel;  // SlaveSelect EIN
}

//-----------------------------------------------------------------------------
void MK3_INIT_FAT16(void) {	// zum initialisieren der SD-Karte
  MK3_DDR_SD|=MK3_SD_SlaveSel;   // SlaveSelect als Ausgang
  MK3_PORT_SD|=MK3_SD_SlaveSel;  // SlaveSelect AUS

  MK3_INIT_SPI();                // INIT der SPI-Ports
}

bool MK3_FAT16_CHKDSK(void)
{	
	bool retvalue = false;
    u8	cnt	= 0;
	struct VBR_Entry *VBR;	
	struct MBR_Entry *MBR;	
	File *file;

	for(cnt=0;cnt<__MAX_FILES_USED;cnt++)
	{
		FilePointer[cnt].state = _UNUSED;	
	}
	
	file = &FilePointer[0];

	SD_Result_t res;
        while(cnt<100) {
          cnt++;
          res=F10_SDC_Init();
          if(res==SD_SUCCESS) break;
        }
        
	

	if(cnt <100)	
	{

	        F12_SDC_GetSector((u32)_MBR_SECTOR,file->buffer);
		MBR = (struct MBR_Entry *) file->buffer;
		FirstPartitionSector = MBR->PartitionEntry1.NoSectorsBeforePartition;
		if((MBR->PartitionEntry1.Type == _FAT16_32_MB_BIOS_Extension) || 
		   (MBR->PartitionEntry1.Type == _FAT16_ST_32_MB) ||
		   (MBR->PartitionEntry1.Type == _FAT16_LT_32_MB))
		{
			F12_SDC_GetSector(FirstPartitionSector,file->buffer);
	
			VBR = (struct VBR_Entry *) file->buffer;
			SectorsPerCluster	= VBR->SectorsPerCluster;
			FatCopies 			= VBR->NoFATCopies;
			PossibleRootEntries = VBR->MaxRootEntries;
			SectorsPerFat 		= VBR->SectorsPerFAT;
			ReservedSectors		= VBR->ReservedSectors;

			FileAllocationTable	=   (u32)(FirstPartitionSector + (u32)ReservedSectors);
			RootDirectory		=   (u32)((u32)FileAllocationTable + (u32)((u32)SectorsPerFat*(u32)FatCopies));
	 		FirstDataCluster	=   (u32)((u32)RootDirectory + ((u32)(PossibleRootEntries>>4)));
			CWD					= 	RootDirectory;
			retvalue = true;
		}

	}
	return(retvalue);	

}

#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1
File * F21_fopen_(cs8 *fname, s8 mode)
{
	File *file;
	
	file = F40_ReserveFilePointer();	
	
	if(file != NULL)	
	{
		file->mode 	= mode;	

		if(F38_SeekFileInDirectory(fname, file))
		{
			if(mode == 'a')		
			{	
				F25_fseek_(file, 0, O_SEEK_END);	
			}
		}		
		else
		{
			if((mode == 'a') || (mode == 'w'))	
			{	
				if(F37_CreateFileInDirectory(fname,file))
				{
					return(file);
				}	
				else
				{
					F41_FreeFilePointer(file);
					file = NULL;
				}
			}
            else     
            {
            	F41_FreeFilePointer(file);			
                file = NULL;              
            }
		}
	}
	return(file);
}
#endif



#if USE_MK3_F_OPENCLOSE==1
void F22_fclose_(File *file)
{

	if(file != NULL)
	{
		I10_fflush_(file);		
	}
	F41_FreeFilePointer(file);		
}
#endif


#if USE_MK3_F_READ==1
u32 F23_fread_(void *buffer, u32 size, u32 count, File *file)
{
	u32 object_cnt 	= 0;				
	u8 *buff_pnt    = 0;		
	u8 success      = 1;			
	
	buff_pnt = (u8 *) buffer;
	
	while((object_cnt < count) && success)
	{		
		while((size > 0) && success)
		{
			*buff_pnt = (u8) F26_fgetchar_(file);
			buff_pnt++;
			size--;
		}
		if(success) object_cnt++;
	}		
	
	return(object_cnt);	
}
#endif


#if USE_MK3_F_WRITE==1
u32 F24_fwrite_(void *buffer, u32 size, u32 count, File *file)
{
	u32 object_cnt 	= 0;	
	u8 *buff_pnt    = 0;			
	u8 success      = 1;		
	
	buff_pnt = (u8 *) buffer;			
	
	while((object_cnt < count) && success)
	{
		while((size > 0) && success)
		{
			success = F27_fputchar_(file, *buff_pnt);
			buff_pnt++;
			size--;
		}
		if(success) object_cnt++;
	}		
	
	return(object_cnt);			
}
#endif									


#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1
s16 F25_fseek_(File *file, s32 offset, Origin_t origin)
{
	s32			fposition 	= 0;
	s16 			retvalue 	= 1;
	u32 	temp 		= 0;
	

	if(origin == O_SEEK_SET)		
	{
		fposition = 0;
	}	
	else if(origin == O_SEEK_END)		
	{
		fposition  = (s32) file->filesize;
	}	
	else if(origin == O_SEEK_CUR)		
	{
		fposition = file->fileposition;
	}	

	fposition += offset;

	if((fposition >= 0) && (fposition <= (s32)file->filesize))
	{
		retvalue 				= 0;
		file->sector_index 		= 0;
		file->byte_index   		= 0;
		file->fileposition 		= 0;
		file->cluster_pointer 	= file->start_cluster;
		
		while(file->fileposition < fposition)
		{
			file->fileposition++;	
			if(file->byte_index < 511)
			{
				file->byte_index++;
			}
			else
			{
				file->byte_index=0;		
				file->sector_index++;	
				if(file->sector_index >= SectorsPerCluster)
				{
					if(file->fileposition < fposition) 
					{
						file->sector_index = 0;
						F31_GetNextCluster(file);
					}
				}
			}	
		}
		if(file->byte_index)				
		{
			temp = (u32)((u32)file->cluster_pointer + (u32)file->sector_index);
			F12_SDC_GetSector((u32)temp,file->buffer);	
		}
	}	
	return(retvalue);
}
#endif


#if USE_MK3_F_READ==1
s16 F26_fgetchar_(File *file)
{	
	s16 c = EOF;
	u32 temp1;
	
	if(((file->fileposition)+1) < file->filesize)
	{
		temp1  = (u32)file->cluster_pointer;
		temp1 += (u32)file->sector_index;
	
		if(file->sector_in_buffer != temp1)
		{
			F12_SDC_GetSector((u32)temp1,file->buffer);
			file->sector_in_buffer = (u32)temp1;
		}	
		c = (s16) file->buffer[file->byte_index];
		file->fileposition++;
	
		if(file->byte_index < 511)	
		{
			file->byte_index++;
		}
		else	
		{
			file->byte_index=0;		
			file->sector_index++;		
			if(file->sector_index >= SectorsPerCluster)	
			{
				file->sector_index = 0;		
				F31_GetNextCluster(file);		
			}
		}
	}
	return(c);
}
#endif

#if USE_MK3_F_WRITE==1
u8 F27_fputchar_(File *file,s8 c)
{	
	u32 ul_temp  = 0;
	u8 retvalue = 1;
	
    if(file->sector_index >= SectorsPerCluster)
	{
		file->sector_index = 0;
		if(!F33_AppendCluster(file)) retvalue = 0;		
	}

	file->buffer[file->byte_index] = c;
	if(file->filesize == file->fileposition) file->filesize++;
	file->fileposition++;	
			
	if(file->byte_index < 511)	
	{
		file->byte_index++;
	}
	else	
	{
		ul_temp  = (u32)file->cluster_pointer;
		ul_temp += (u32)file->sector_index;
		
		F11_SDC_PutSector((u32)ul_temp,file->buffer);
		file->byte_index=0;
		file->sector_index++;
		if(file->sector_index >= SectorsPerCluster)	
		{
			file->sector_index = 0;
			if(!F33_AppendCluster(file)) retvalue = 0;		
		}
	}
	return(retvalue);
}
#endif

#if USE_MK3_F_WRITE==1
u8 F28_fputs_(File *file,cs8 * string)
{
	u8 i=0;
	
	while(string[i] != 0)
	{
		F27_fputchar_(file,string[i]);
		i++;
	}
	return(0);
}
#endif

#if USE_MK3_F_READ==1
s8 * F29_fgets_(s8 *string, s16 count, File *file)
{
	s16 buff_pnt = 0;
	s16 buff_tmp = 0;
	s8 *retvalue = 0;
	
	retvalue = string;

	while(count > 1)	
	{
		buff_tmp = F26_fgetchar_(file);	
		if(buff_tmp == EOF) 	
		{
			if(buff_pnt != 0)	
			{
				break;
			}
			else
			{
				retvalue = NULL;		
				break;
			}
		}	
		else if(buff_tmp == 0x0A)	
		{		
			string[buff_pnt] = (s8) buff_tmp;	
			buff_pnt++;
			break;
		}
		else
		{
			string[buff_pnt] = (s8) buff_tmp;
			count--;
			buff_pnt++;
		}
	}
	string[buff_pnt] = 0;	
	return(retvalue);
}
#endif

#if USE_MK3_F_SEARCH==1
u8 F30_fexist_(cs8 *fname)
{
	File *file;

	file = F21_fopen_(fname,'r');	
	
	if(file != NULL)
	{
		F41_FreeFilePointer(file); 
		return(1);			
	}
	else
	{
		return(0);	
	}
}
#endif


#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_READ==1 || USE_MK3_F_SEARCH==1 || USE_MK3_F_DIR==1 || USE_MK3_F_FIND==1 || USE_MK3_F_DELETE==1
u16 F31_GetNextCluster(File *file)
{
	u32 fat_pointer			= 0;
	u32 fat_sector_offset	        = 0;
	u32 ul_tmp			= 0;
	u8 retvalue			= 0;	
	
	
	if((file->cluster_pointer >= RootDirectory) && (file->cluster_pointer < (RootDirectory + 31)))
	{							
		file->cluster_pointer++;		
		retvalue = 1;						
	}
	else if(file->cluster_pointer > (RootDirectory + 31))	
	{
		fat_sector_offset   = ((file->cluster_pointer) - (FirstDataCluster));	
		fat_sector_offset  /= SectorsPerCluster;
		fat_sector_offset  += 2;	
		fat_pointer = (fat_sector_offset%0x100);	
		fat_sector_offset   = (fat_sector_offset>>8);	

		F12_SDC_GetSector((u32)(FileAllocationTable + fat_sector_offset),file->buffer);
		file->sector_in_buffer = (FileAllocationTable + fat_sector_offset);
	
		ul_tmp  = (u32)file->buffer[((fat_pointer << 1)+1)];
		ul_tmp  = (ul_tmp << 8);
		ul_tmp |= (u32)file->buffer[(fat_pointer << 1)];   
		ul_tmp -=2;			
		ul_tmp *= SectorsPerCluster;		
		ul_tmp += FirstDataCluster;		
	
		if(ul_tmp < 0xfff7)	
		{
			file->cluster_pointer = (u32) ul_tmp;	
			retvalue = 1;		
		}	
	}
	return(retvalue);
}
#endif

#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_WRITE==1 || USE_MK3_F_SEARCH==1 || USE_MK3_F_DIR==1
u16 F32_FindNextFreeCluster(File *file)
{
	u32 fat_pointer			= 0;
	u32 ul_tmp			= 0;
	u16  fat_sector_offset	        = 0;
	u16  fat_entry			= 0;
	u16  free_cluster		= 0;
	
	fat_pointer = (u32) FileAllocationTable;
							
	while((fat_sector_offset < SectorsPerFat) && (!free_cluster))	
	{
		ul_tmp = (u32) ((u32)fat_pointer + (u32)fat_sector_offset);
		F12_SDC_GetSector((u32)ul_tmp,file->buffer);
		file->sector_in_buffer = ul_tmp;
		Fat = (struct FatEntry *)file->buffer;
		for(fat_entry=0;fat_entry<256;fat_entry++)
		{
			if(Fat[fat_entry].next_cluster == 0x0000)
			{				
				Fat[fat_entry].next_cluster = 0xffff;
				F11_SDC_PutSector((u32)file->sector_in_buffer,file->buffer);
				free_cluster  = fat_entry;
				free_cluster += (fat_sector_offset << 8);
				fat_entry = 256;
			}
		}
		fat_sector_offset++;
	}	
return(free_cluster);
}
#endif


#if USE_MK3_F_WRITE==1
u8 F33_AppendCluster(File *file)
{
	u16  free_cluster = 0;	
	u32 fat_pointer	= 0;
	u8 retvalue    	= 0;
	
	free_cluster = F32_FindNextFreeCluster(file);
	if(free_cluster) retvalue = 1;	
	fat_pointer  = FileAllocationTable;	
	fat_pointer += (u32)((u32)F35_GetFatClusterOffset(file) >> 8);
	
        F12_SDC_GetSector(fat_pointer,file->buffer);       
	Fat = (struct FatEntry *)file->buffer;
	Fat[F36_GetFatSectorIndex(file)].next_cluster = free_cluster;
	F11_SDC_PutSector((u32)fat_pointer,file->buffer);		

	fat_pointer  = (u32)free_cluster;
	fat_pointer -= 2;
	fat_pointer *= SectorsPerCluster;
	fat_pointer += FirstDataCluster;

	file->cluster_pointer = fat_pointer;	
	return(retvalue);	
}
#endif


#if USE_MK3_F_DELETE==1
void F34_DeleteClusterChain(u16 startcluster)
{
	u16 fat_index			= 0;
	u16 fat_sector_offset	        = 0;
	u32 sector_in_buffer	        = 0;	
	u32 ul_temp			= 0;
	u8 buffer[512];
	
	
	fat_index = (startcluster % 0x100);		
	fat_sector_offset = (startcluster >> 8);		
	ul_temp = (u32)(FileAllocationTable + fat_sector_offset);
		
	
	do
	{
		if(sector_in_buffer != ul_temp)
		{
			sector_in_buffer = ul_temp;
			F12_SDC_GetSector(ul_temp,buffer);	
		}
		Fat = (struct FatEntry *)buffer; 
	
		startcluster = Fat[fat_index].next_cluster;
		Fat[fat_index].next_cluster = 0x0000;

		if(startcluster != 0xffff)
		{
			fat_index = (startcluster % 0x100);	
			fat_sector_offset = (startcluster >> 8);
			ul_temp = (u32)(FileAllocationTable + fat_sector_offset);
			if((sector_in_buffer != ul_temp) || (startcluster == 0xffff))
			{
				F11_SDC_PutSector(sector_in_buffer,buffer);	
			}
		}
	}
	while(startcluster != 0xffff);
}
#endif


#if USE_MK3_F_WRITE==1
u16 F35_GetFatClusterOffset(File *file)
{
	u32 fat_sector_offset	=	0;

	fat_sector_offset   = ((file->cluster_pointer) - (FirstDataCluster));
	fat_sector_offset  /= SectorsPerCluster;
	fat_sector_offset  += 2;	
	
	return((u16)fat_sector_offset);
}
#endif

#if USE_MK3_F_WRITE==1
u16 F36_GetFatSectorIndex(File *file)
{
	u16 fat_pointer			=	0;
	
	fat_pointer = F35_GetFatClusterOffset(file);
	fat_pointer = fat_pointer % 0x100;
	
	return(fat_pointer);
}
#endif

#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1
u8 F37_CreateFileInDirectory(cs8 *fname, File *file)
{
	u16 	rootentry	= 0;	
	u16 	cnt_enries_searched = 0;	
	u8 		i = 0;
	u16 	sector_offset = 0;	
	u8 		retvalue = 0;
	u32		cluster_temp = 0;
	u16		cluster = 0;
	s8 name[11];

	name[0]=0;	

	F39_SeperateFileName(fname,name);

	cluster_temp = (u32)F32_FindNextFreeCluster(file);	
	
	if(cluster_temp) 	
	{
		cluster = (u16)cluster_temp;	
		cluster_temp -=2;	
		cluster_temp *= SectorsPerCluster;	
		file->start_cluster   = (FirstDataCluster + cluster_temp);
		file->cluster_pointer = file->start_cluster;
	
			
		do
		{
			rootentry=0;		
			F12_SDC_GetSector((u32)(CWD + sector_offset),file->buffer);
			DirectoryEntry = (struct DirEntry *)file->buffer;
			while((rootentry<16) && (!retvalue))
			{
				if((DirectoryEntry[rootentry].attribute == 0) || (DirectoryEntry[rootentry].attribute == 0xE5))	
				{
					for(i=0;i<11;i++) DirectoryEntry[rootentry].name[i] = name[i];
					DirectoryEntry[rootentry].attribute    = _FILE;
					DirectoryEntry[rootentry].startcluster = cluster;
					DirectoryEntry[rootentry].size     = 0;
					file->directory_sector = (u32) (CWD + sector_offset);
					file->directory_index  = (u8) rootentry;
					retvalue = 1;
					F11_SDC_PutSector((u32)(CWD + sector_offset),file->buffer);
				}			
				rootentry++;
				cnt_enries_searched++;
			}
			if(!retvalue)
			{
				rootentry = 0;
				sector_offset++;
			}
		}
		while((cnt_enries_searched< PossibleRootEntries) && (!retvalue));

	}
	return(retvalue);	
}
#endif


#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1
u8 F38_SeekFileInDirectory(cs8 *fname, File *file)
{
	u16 	rootentry=0;
	u16   	end_of_directory_not_reached = 0;	
	u8 		i=0;
	u8  	retvalue = 0;
	u32  	cluster_temp = 0;
	s8 name[13];

	name[0]=0;			
	
	F39_SeperateFileName(fname,name);
	
	file->cluster_pointer = CWD;			

				
	do
	{		
		rootentry=0;			
		F12_SDC_GetSector((u32) file->cluster_pointer,file->buffer);
		DirectoryEntry = (struct DirEntry *)file->buffer;

		while((!retvalue)&&(rootentry<16))
		{
			i=0;
			if(DirectoryEntry[rootentry].name[0] != 0xe5)
			{
				while((i<=10)&&(DirectoryEntry[rootentry].name[i] == name[i]))
				{
					i++;
				}
			}

			if((i==11) && (DirectoryEntry[rootentry].attribute & _FILE))
			{
				cluster_temp = (u32)DirectoryEntry[rootentry].startcluster;
				cluster_temp -=2;
				cluster_temp *= (u32)SectorsPerCluster;
				file->start_cluster   = (FirstDataCluster + cluster_temp);
				file->directory_sector = (u32) file->cluster_pointer;
				file->cluster_pointer = file->start_cluster;
				file->filesize = (u32) DirectoryEntry[rootentry].size;
				file->directory_index  = (u8) rootentry;
				retvalue = 1;
			}
			rootentry++;
		}
		if(!retvalue)
		{
			end_of_directory_not_reached = F31_GetNextCluster(file);
		}
	}
	while((end_of_directory_not_reached) && (!retvalue));
	return(retvalue);
}
#endif

#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1 || USE_MK3_F_DIR==1 || USE_MK3_F_FIND==1 || USE_MK3_F_DELETE==1
void F39_SeperateFileName(cs8 *fname, s8 *name)
{
	u8 readpointer	= 0;
	u8 writepointer	= 0;
	u8 attribute	= 1;
	u8 i = 0;
	
	
	while((writepointer<=10) && (fname[readpointer]!=0))
	{					
		if(fname[readpointer]=='.')	
		{
			if(attribute)		
			{
				name[writepointer] = fname[readpointer];
				readpointer++;
				writepointer++;			
			}
			else
			{
				if(fname[(readpointer-1)] != '*')
				{
					for(i=writepointer;i<8;i++)
					{
						name[i] = ' ';
					}
				}
				readpointer++;
				writepointer = 8;
			}
		}
		else if(fname[readpointer] == '*')
		{
			if(writepointer < 8)		
			{
				readpointer++;
				writepointer = 8;	
			}
			else			
			{
				writepointer = 11;	
			}
			attribute = 0;
		}
		else
		{
			if((fname[readpointer]>96) && (fname[readpointer]<123))
			{
				name[writepointer]=(fname[readpointer] - 32);
			}
			else
			{
				name[writepointer]=fname[readpointer];
			}	
			readpointer++;
			writepointer++;
			attribute = 0;
		}
	}	
}
#endif

#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1
File * F40_ReserveFilePointer(void)
{
	File *file;
	file = NULL;
	u8	temp;
	
	for(temp = 0;temp<__MAX_FILES_USED;temp++)
	{
		if(FilePointer[temp].state == _UNUSED)
		{
			file = &FilePointer[temp];
			FilePointer[temp].state 			= _USED;
			FilePointer[temp].mode 				= 0;
			FilePointer[temp].start_cluster			= 0;
			FilePointer[temp].cluster_pointer		= 0;
			FilePointer[temp].sector_index 			= 0;
			FilePointer[temp].byte_index 			= 0;
			FilePointer[temp].filesize 			= 0;
			FilePointer[temp].fileposition 			= 0;
			FilePointer[temp].sector_in_buffer 		= 0;
			FilePointer[temp].directory_sector 		= 0;
			FilePointer[temp].directory_index 		= 0;
			FilePointer[temp].attribute 			= 0;
			break;
		}
	}
	return(file);
}
#endif

#if USE_MK3_F_OPENCLOSE==1 || USE_MK3_F_SEARCH==1
void F41_FreeFilePointer(File *file)
{
	u8 cnt = 0;
	
	for(cnt=0;cnt<__MAX_FILES_USED;cnt++)
	{
		if(&FilePointer[cnt] == file)
		{
			FilePointer[cnt].state = _UNUSED;	
		}
	}
}
#endif


#if USE_MK3_F_DELETE==1
void F42_DeleteDirectoryEntry(Find *item)
{
	u8 buffer[512];
	
	
	F12_SDC_GetSector((u32) item->cluster_pointer,buffer);
	DirectoryEntry = (struct DirEntry *)buffer;

	DirectoryEntry[(item->directory_index)-1].attribute = 0;
	DirectoryEntry[(item->directory_index)-1].name[0] = 0xE5;
	F11_SDC_PutSector((u32) item->cluster_pointer,buffer);
}
#endif



#if USE_MK3_F_DIR==1
u8 F43_CreateSubDirectory(cs8 *fname)
{
	u16 	index	= 0;	
	u16 	cnt_entries_searched = 0;
	u16		i = 0;
	u16 	sector_offset = 0;
	u8 		retvalue = 0;
	u32		cluster_temp = 0;
	u16		cluster = 0;
	File 	file;
	s8		name[11];

	name[0]=0;

	F39_SeperateFileName(fname,name);
	cluster_temp = (u32)F32_FindNextFreeCluster(&file);
	
	if(cluster_temp) 		
	{
		cluster = (u16)cluster_temp;	
		cluster_temp -=2;	
		cluster_temp *= SectorsPerCluster;		
		file.start_cluster   = (FirstDataCluster + cluster_temp);
		file.cluster_pointer = file.start_cluster;	
	
		for(i=0;i<512;i++)
		{
			file.buffer[i] = 0;	
		}
		for(sector_offset=0;sector_offset<SectorsPerCluster;sector_offset++)
		{
			F11_SDC_PutSector((u32)(file.start_cluster + sector_offset),file.buffer);
		}
                
		DirectoryEntry = (struct DirEntry *)file.buffer;
		DirectoryEntry[0].name[0] = '.';
		DirectoryEntry[0].attribute = _DIRECTORY;
		DirectoryEntry[0].startcluster = cluster;

		DirectoryEntry[1].name[0] = '.';
		DirectoryEntry[1].name[1] = '.';	
		DirectoryEntry[1].attribute = _DIRECTORY;
		if(CWD == RootDirectory)
		{
			DirectoryEntry[1].startcluster = 0;	
		}
		else
		{
			cluster_temp = (CWD - FirstDataCluster);
			cluster_temp /= SectorsPerCluster;
			cluster_temp -= 2;
			DirectoryEntry[1].startcluster = (u16) cluster_temp;
		}
		F11_SDC_PutSector((u32) file.start_cluster,file.buffer);

		sector_offset = 0;
		cnt_entries_searched = 0;
		do
		{			
			index=0;	
			F12_SDC_GetSector((u32)(CWD + sector_offset),file.buffer);
			DirectoryEntry = (struct DirEntry *)file.buffer;
			while((index<16) && (!retvalue))
			{
				if((DirectoryEntry[index].attribute == 0) || (DirectoryEntry[index].attribute == 0xE5))
				{
					for(i=0;i<11;i++) DirectoryEntry[index].name[i] = name[i];
					DirectoryEntry[index].attribute    = _DIRECTORY;
					DirectoryEntry[index].startcluster = cluster;
					DirectoryEntry[index].size     = 0;
					file.directory_sector = (u32) (CWD + sector_offset);
					file.directory_index  = (u8) index;
					retvalue = 1;
					F11_SDC_PutSector((u32)(CWD + sector_offset),file.buffer);
				}			
				index++;
				cnt_entries_searched++;
			}
			if(!retvalue)	
			{
				index = 0;
				sector_offset++;
			}
		}
		while((cnt_entries_searched< PossibleRootEntries) && (!retvalue));

	}
	return(retvalue);		
}
#endif


#if USE_MK3_F_DIR==1
u16 F44_SeekSubDirectory(cs8 *fname)
{
	u16 	index = 0;
	u16   	end_of_directory_not_reached = 0;	
	u8 		i = 0;
	u16  	cluster_temp = 0;
	s8 name[11];
	File 	file;

	name[0]=0;

	F39_SeperateFileName(fname,name);
	
	file.cluster_pointer = CWD;			
	file.start_cluster   = CWD;	

					
	do
	{				
		index=0;		
		F12_SDC_GetSector((u32) file.cluster_pointer,file.buffer);
		DirectoryEntry = (struct DirEntry *)file.buffer;

		while((!cluster_temp)&&(index<16))
		{
			i=0;
			if(DirectoryEntry[index].name[0] != 0xe5)
			{
				while((i<=10)&&(DirectoryEntry[index].name[i] == name[i]))
				{
					i++;
				}
			}

			if((i==11) && (DirectoryEntry[index].attribute & _DIRECTORY))
			{
				cluster_temp = (u16)DirectoryEntry[index].startcluster;
			}
			index++;
		}
		if(!cluster_temp)		
		{
			end_of_directory_not_reached = F31_GetNextCluster(&file);
		}
	}
	while((end_of_directory_not_reached) && (!cluster_temp));
	return(cluster_temp);
}
#endif


#if USE_MK3_F_DIR==1
u8	F45_mkdir_(cs8 *fname)
{
	u8 retvalue = 0;
	
	retvalue = F44_SeekSubDirectory(fname);
	
	if(!retvalue)
	{
		F43_CreateSubDirectory(fname);	
		retvalue = 1;
	}
	else
	{
		retvalue = 0;
	}
	
	return(retvalue);
}
#endif


#if USE_MK3_F_DIR==1
u8	F46_chdir_(cs8 *fname) 
{
	u8 	retvalue = 0;
	s8  name[11];
	
	u32 ultemp = 0;

	name[0]=0;

	F39_SeperateFileName(fname,name);	
	
	ultemp = (u32)F44_SeekSubDirectory(name);
	if(ultemp >= 2)
	{
		ultemp -=2;		
		ultemp *= SectorsPerCluster;
		ultemp += FirstDataCluster;
		CWD = ultemp;
		retvalue = 1;
	}
	else
	{
		CWD = RootDirectory;
		retvalue = 1;
	}

	return(retvalue);
}
#endif

#if USE_MK3_F_FIND==1 || USE_MK3_F_DELETE==1
u8 F47_FindItem(Find *item)
{
	u16 	index = 0;
	u16   	end_of_directory_not_reached = 0;	
	u8 		i = 0;
	u8		readpointer = 0;
	u8		writepointer = 0;
	u8		retvalue = 0;
	File 	file;

	
	file.cluster_pointer = item->cluster_pointer;
	file.start_cluster   = item->cluster_pointer;	
	index 				 = item->directory_index;
								
	do
	{									
		F12_SDC_GetSector((u32) file.cluster_pointer,file.buffer);	
		DirectoryEntry = (struct DirEntry *)file.buffer;

		while((!retvalue)&&(index<16))
		{
			i=0;			
			if(DirectoryEntry[index].name[0] != 0xe5)
			{
				while((i<=10)&&((DirectoryEntry[index].name[i] == item->searchstring[i]) || (item->searchstring[i]=='*') || item->searchstring[i]=='?'))
				{
					i++;
				}
			}

			if((i==11) && (DirectoryEntry[index].attribute & item->attribute))
			{
				for(readpointer=0;readpointer<=10;readpointer++)
				{
					if(DirectoryEntry[index].name[readpointer] != ' ')
					{
						item->name[writepointer] = DirectoryEntry[index].name[readpointer];	
						writepointer++;
					}
					else if((readpointer==7) && (DirectoryEntry[index].attribute == _FILE))
					{
						if(DirectoryEntry[index].name[readpointer] != ' ')
						{
							item->name[writepointer] = DirectoryEntry[index].name[readpointer];	
							writepointer++;
						}
						item->name[writepointer] = '.';
						writepointer++;
					}
				}
				item->startcluster = (u16)DirectoryEntry[index].startcluster;
				item->directory_index = ++index;
				item->cluster_pointer = file.cluster_pointer;
				retvalue = 1;
			}
			index++;
		}
		if(!retvalue)				
		{
			end_of_directory_not_reached = F31_GetNextCluster(&file);
		}
		index = 0;
	}
	while((end_of_directory_not_reached) && (!retvalue));
	
	return(retvalue);	
}
#endif


#if USE_MK3_F_FIND==1 || USE_MK3_F_DELETE==1
u8 F48_findfirst_(cs8 *fname, Find *item, u8 attribute)
{
	u8 retvalue = 0;
	u8 i = 0;

	for(i=0;i<=11;i++)
	{
		item->searchstring[i] = '*';
		item->name[i] = 0;
	}

	F39_SeperateFileName(fname,item->searchstring);
	
	item->cluster_pointer = CWD;		
	item->directory_index = 0;
	item->attribute = attribute;
	
	retvalue = F47_FindItem(item);

	return(retvalue);	
}
#endif

#if USE_MK3_F_FIND==1 || USE_MK3_F_DELETE==1
u8 F49_findnext_(Find *item)
{
	u8 retvalue = 0;
	u8 i = 0;

	for(i=0;i<=11;i++)			
	{
		item->name[i] = 0;
	}

	retvalue = F47_FindItem(item);			

	return(retvalue);	
}
#endif


#if USE_MK3_F_DELETE==1
u8 F50_fdelete_(cs8 *fname)
{
	u8 retvalue = 0;
	Find item;
	
	retvalue = F48_findfirst_(fname,&item, _FILE);	

	if(retvalue);				
	{
		if(item.startcluster)	
		{
			F34_DeleteClusterChain(item.startcluster);		
		}
		F42_DeleteDirectoryEntry(&item);	
	
		do
		{
			retvalue = F49_findnext_(&item);
			if(retvalue)
			{
				F34_DeleteClusterChain(item.startcluster);
				F42_DeleteDirectoryEntry(&item);	
			}
		}
		while(retvalue);	
	}

	return(retvalue);	
}
#endif

#if USE_MK3_F_DELETE==1
u8 F51_rmdir_(cs8 *dname)
{
	u8 retvalue = 0;
	Find item;
	
	retvalue = F48_findfirst_(dname,&item, _DIRECTORY);		

	if(retvalue);						
	{
		F34_DeleteClusterChain(item.startcluster);		
		F42_DeleteDirectoryEntry(&item);	
	
		do
		{
			retvalue = F49_findnext_(&item);
			if(retvalue)
			{
				F34_DeleteClusterChain(item.startcluster);
				F42_DeleteDirectoryEntry(&item);
			}
		}
		while(retvalue);	
	}

	return(retvalue);	
}
#endif

#if USE_MK3_F_READ==1
u16 F52_feof_(File *file)
{	
	if(((file->fileposition)+1) < (file->filesize))
	{
		return(0);
	}
	else 
	{
		return(1);	
	}
}
#endif
